#!/usr/bin/env python3
#
# Pucktada Treeratpituk (https://pucktada.github.io/)
# License: MIT
# 2017-05-01
#
# A set of utility functions for reading and writing to TFRecord

import tensorflow as tf
import numpy as np
from os import listdir
from os.path import isfile, isdir, join

def get_num_records(filename):
    """ return the number of record within a given TFRecord file """
    return len([x for x in tf.python_io.tf_record_iterator(filename)])
    
# Write all examples into a TFRecords file
def save_tfrecord(writer, sources, targets):
    """ write the sources and targets sequences into a TFRecord file (given by writer) """
    for source, target in zip(sources, targets):
        ex = make_example(source, target)
        writer.write(ex.SerializeToString())

def make_example(source, target):
    """ create a SequenceExample out of a sequence of source and target """
    # The object we return
    ex = tf.train.SequenceExample()
    # A non-sequential feature of our example
    sequence_length = len(source)
    ex.context.feature["length"].int64_list.value.append(sequence_length)
    
    # Feature lists for the two sequential features of our example
    fl_source = ex.feature_lists.feature_list["source"]
    fl_target = ex.feature_lists.feature_list["target"]
    for src, tar in zip(source, target):
        #print(type(token))
        fl_source.feature.add().int64_list.value.append(src)
        fl_target.feature.add().int64_list.value.append(tar)
    return ex

def read_and_decode_single_example(filenames, shuffle=False, num_epochs=None):
    # first construct a queue containing a list of filenames.
    # this lets a user split up there dataset in multiple files to keep size down
    #filename_queue = tf.train.string_input_producer([filename], num_epochs=10)
    filename_queue = tf.train.string_input_producer(filenames, 
        shuffle=shuffle, num_epochs=num_epochs)
    
    reader = tf.TFRecordReader()
    # One can read a single serialized example from a filename
    # serialized_example is a Tensor of type string.
    key, serialized_ex = reader.read(filename_queue)
    context, sequences = tf.parse_single_sequence_example(serialized_ex,
            context_features = {
                "length": tf.FixedLenFeature([], dtype=tf.int64)
            },
            sequence_features={
                # We know the length of both fields. If not the
                # tf.VarLenFeature could be used
                "source": tf.FixedLenSequenceFeature([], dtype=tf.int64),
                "target": tf.FixedLenSequenceFeature([], dtype=tf.int64)
            })
    return (key, context, sequences)
        
def test_sequence_handler():
    print('testing sequence_handler')
    tmp_filename = 'tf.tmp'

    writer = tf.python_io.TFRecordWriter(tmp_filename)
    xlist = [[1, 2, 3], [1, 2], [3, 2, 1], [5,6,7,8,9,9,8], [7,6,6,7]]
    ylist = [[0, 1, 0], [2, 0], [2, 2, 2], [3,1,3,1,3,1,1], [4,1,4,1]]
    print(xlist)
    print(ylist)
    save_tfrecord(writer, xlist, ylist)
    writer.close()
    
    graph = tf.Graph()
    with graph.as_default():
        key, context, features = read_and_decode_single_example([tmp_filename])
        tokens = features['source']
        labels = features['target']
        
    with tf.Session(graph=graph) as sess:
        init_op = tf.group(tf.global_variables_initializer(), 
                           tf.local_variables_initializer())
        sess.run(init_op)
    
        coord = tf.train.Coordinator()
        threads = tf.train.start_queue_runners(coord=coord)

        print('\nread back the file...')
        for i in range(len(xlist)):
            t, l = sess.run([tokens, labels])
            print('tokens:', t)
            print('labels:', l)
            print('')            

        # stop our queue threads and properly close the session
        coord.request_stop()
        coord.join(threads)
  
def count_sentence(dirname):
    files = [join(dirname, f) for f in listdir(dirname)]
    
    total = 0
    for f in files:
        num = get_num_records(f)
        total += num
        
    print('total', total)
        
if __name__ == '__main__':
    #test_sequence_handler()
    count_sentence('data/train')
    
    


#!/usr/bin/env python
import math
from os.path import join
from tensorflow.python.lib.io import file_io
import sys
import re
from random import randint
import configargparse
import logging
import tensorflow as tf
import numpy as np
from char_dictionary import CharDictionary
from cutkum import CkModel
from sequence_handler import save_tfrecord, read_and_decode_single_example, get_num_records
import util as util

def get_number_of_records(dirpath):
    ''' get the number of records for all tf files inside a dirpath '''
    train_files = [join(dirpath, f) for f in file_io.list_directory(dirpath)]
    
    num_records = 0
    for filename in train_files:
        num_records += get_num_records(filename)
    return num_records

def validate(sess, model, char_dict, one_hot_by_t, seq_lengths, chars_mat, boundary_mat):
    feed = { model.inputs: one_hot_by_t, 
        model.seq_lengths: seq_lengths, 
        model.keep_prob: 1.0
    }
    #print('run_session', one_hot_by_t.shape)
    probs = sess.run(model.probs, feed_dict=feed)
    #print('DONE run session')

    total_c = 0.0
    total_r = 0.0
    total_o = 0.0
    _, n_examples, _ = probs.shape
    #print('n_examples', n_examples)

    for i in range(n_examples):
        p = probs[0:seq_lengths[i], i, :]
        labels = util.viterbi(p)
        words  = char_dict.chars2words(chars_mat[i], labels)

        x = boundary_mat[i]
        y = util.get_boundary_array(words)
        
        n_answers = util.word_count_from_boundary_array(x)
        n_words   = len(words)
        correct, n_refwords, n_outwords = util.count_correct_words(x, y, n_answers, n_words)

        total_c += correct
        total_r += n_refwords
        total_o += n_outwords

    r = (float)(total_c / total_r)
    p = (float)(total_c / total_o)
    f = 2 * p * r / (p + r)
    return r, p, f

# [article, encyclopedia, news, novel] ==> 148995 sentences
def train(model_settings, train_settings):
    train_files = [join(train_settings['input_dir'], f) for f in file_io.list_directory(train_settings['input_dir'])]
    logging.info('#training_files %d', len(train_files))
    test_files = [join(train_settings['eval_dir'], f) for f in file_io.list_directory(train_settings['eval_dir'])]
    logging.info('#testing_files %d', len(test_files))
    
    char_dict = CharDictionary()
    model_settings['input_classes'] = char_dict.num_char_classes() + 1
    model_settings['label_classes'] = char_dict.num_label_classes() + 1

    test_one_hot_by_t, test_seq_lengths, test_chars_mat, test_boundary_mat = util.load_validation_set(test_files)
    
    bucket_boundaries = [20,40,80,120,160,200,240,280,320,360,400,500,600]
    graph = tf.Graph()
    recall, precision, fmeasure = 0, 0, 0
    with graph.as_default():
        with tf.device('/gpu:0'):
            model = CkModel(model_settings)
            model.build_graph()
        
        key, context, sequences = read_and_decode_single_example(train_files, shuffle=True)
        length = tf.cast(context["length"], tf.int32)
        src = sequences['source']
        tar = sequences['target']
        
        batch_lengths, [batch_src, batch_tar] = tf.contrib.training.bucket_by_sequence_length(
            input_length=length,
            tensors=[src, tar],
            batch_size=train_settings['batch_size'],
            bucket_boundaries=bucket_boundaries,
            dynamic_pad=True,
            name='batch'
        )

        in_embedding  = np.eye(model.input_classes)
        one_hot_src   = tf.nn.embedding_lookup(in_embedding, batch_src)
        
        out_embedding = np.eye(model.label_classes)
        one_hot_tar   = tf.nn.embedding_lookup(out_embedding, batch_tar)

        train_ops = [model.train_op, model.summary_op, model.mean_loss, model.num_entries, model.probs]
        
    with tf.Session(graph=graph, config=tf.ConfigProto(allow_soft_placement=True)) as sess:
        init_op = tf.group(tf.global_variables_initializer(), 
                            tf.local_variables_initializer())
        sess.run(init_op)
        
        coord = tf.train.Coordinator()
        threads = tf.train.start_queue_runners(sess, coord=coord)
        
    	writer = tf.summary.FileWriter(train_settings['job_dir'] + "/log/" + train_settings['job_name'])
        writer.add_graph(sess.graph)
        
        saver = tf.train.Saver(max_to_keep=3, keep_checkpoint_every_n_hours=5) 
        ckpt = tf.train.get_checkpoint_state(train_settings['job_dir'])
        if ckpt and ckpt.model_checkpoint_path:
            logging.info('loading checkpoint: %s', ckpt.model_checkpoint_path)
            saver.restore(sess, ckpt.model_checkpoint_path)
        
        step = sess.run(model.global_step) # load step from checkpoint
        
        #for step in range(train_settings['num_step']):
        while step < train_settings['num_step']:
            _len, _in, _out, _src, _tar = sess.run([batch_lengths, batch_src, batch_tar, 
                one_hot_src, one_hot_tar])
            src_by_times = np.stack(_src, axis=1)
            tar_by_times = np.stack(_tar, axis=1)

            offset = randint(0, min(model_settings['num_unroll'], max(_len), 5)-1)

            max_length = src_by_times.shape[0]
            num_rolls = (int)(math.ceil((max_length - offset) / float(model_settings['num_unroll'])))
            total_loss    = 0
            total_entries = 0

            probs  = np.zeros(shape=[0, train_settings['batch_size'], model.label_classes])
            
            len_left   = _len - offset

            padded_len = np.full(src_by_times.shape[1], model_settings['num_unroll'])

            #for i in range(1):
            for i in range(num_rolls):
                start_t = i * model_settings['num_unroll'] + offset
                end_t   = min((i + 1) * model_settings['num_unroll'] + offset, max_length)

                roll_len = np.minimum.reduce([len_left, padded_len])
                roll_len[roll_len < 0] = 0
                len_left = len_left - roll_len

                feed = { model.inputs:  src_by_times[start_t:end_t,:,:],
                         model.outputs: tar_by_times[start_t:end_t,:,:],
                         #model.seq_lengths: _len,
                         model.seq_lengths: roll_len,
                         model.keep_prob: model_settings['keep_prob'] }
                _, _summary, _mean_loss, _num_entries, _p = sess.run(train_ops, feed_dict=feed)

                probs  = np.concatenate((probs, _p), axis=0)
                total_loss += (_mean_loss * _num_entries)
                total_entries += _num_entries
            
            step, _ = sess.run([model.global_step, model.increment_global_step_op])

            if (step % 20 == 0) and (total_entries != 0):
                #print('saving ckpoint...')
                saver.save(sess, train_settings['job_dir'] + '/' + train_settings['job_name'], 
                    global_step=model.global_step)
                avg_loss = total_loss / total_entries
                
                writer.add_summary(_summary, global_step=step)
                print("step: %6d, loss: %f" % (step, avg_loss))

                if step % 500 == 0:

                    #mask = np.sign(_in)
                    #mask = mask[:, offset:]
                    #results = np.argmax(probs, axis=2)
                    #results = np.stack(results, axis=1)

                    #results = results * mask
                    
                    #_cids   = _in[0, offset:]
                    #_labels = _out[0, offset:]
                    #_pred   = results[0,:]
                    
                    #_chars  = char_dict.cids2chars(_cids)
                    #_words  = char_dict.chars2words(_chars, _labels)
                    #_pwords = char_dict.chars2words(_chars, _pred)
                    #print('|'.join(_words))
                    #print('')
                    #print('|'.join(_pwords))

                    recall, precision, f = validate(sess, model, char_dict, test_one_hot_by_t, 
                    test_seq_lengths, test_chars_mat, test_boundary_mat)
                    pr_summary = tf.Summary(value=[
                        tf.Summary.Value(tag="recall", simple_value=recall),                    
                        tf.Summary.Value(tag="precision", simple_value=precision),
                        tf.Summary.Value(tag="F-measure", simple_value=f)
                    ])
                    writer.add_summary(pr_summary, global_step=step)
        
        coord.request_stop()
        coord.join(threads)

if __name__ == '__main__':
    logging.info('train word segmentor model')

    p = configargparse.getArgParser()
    p.add('-c', '--config', required=False, is_config_file=True, help='config file path')
    p.add('-v', '--verbose', help='verbose', action='store_true')
    p.add('-d', '--input-dir', required=True, help='input directory of tfrecord files for training')
    p.add('-e', '--eval-dir',  required=True, help='directory of text files for cross validation during training')
    p.add('-o', '--job-dir', required=True, help='output directory to save')
    p.add('-j', '--job-name', required=True, help='job name (used in saving model and logging)')
    
    #p.add('-l', '--logdir', required=True, help='log directory for tensorboard')
    p.add('--cell-size',  required=True, help='number of cell states', type=int, action='append')
    p.add('--num-layers', required=True, help='number of layers', type=int)
    p.add('--keep-prob', required=True, help='keep probability (for dropout)', type=float)
    
    #p.add('--num-unroll', required=True, help='number of unroll for rnn', type=int)
    p.add('--learning-rate', required=True, help='initial learning rate', type=float)
    p.add('--l2-regularization', required=True, help='l2 regularization factor', type=float)
    p.add('--num-step', required=True, help='number of steps to train', type=int)
    p.add('--batch-size', required=True, help='batch size', type=int)

    cell_type = p.add_mutually_exclusive_group(required=True)
    cell_type.add('--rnn',  help='use RNN cell',  dest='cell', action='store_const', const='rnn')
    cell_type.add('--lstm', help='use LSTM cell', dest='cell', action='store_const', const='lstm')
    #cell_type.add('--gru',  help='use GRU cell',  dest='cell', action='store_const', const='gru')
    #group.set_defaults(cell='rnn')
    
    direction = p.add_mutually_exclusive_group(required=True)
    direction.add('--fw', help='only forward layer',   dest='direction', action='store_const', const=1)
    direction.add('--bi', help='bi-directional layer', dest='direction', action='store_const', const=2)

    opts = vars(p.parse_known_args()[0])
    verbose = opts['verbose']
    assert(len(opts['cell_size']) == opts['num_layers'])
    print(opts)

    # MODEL
    model_settings = dict()
    model_settings['num_layers'] = opts['num_layers']
    model_settings['cell_size']  = opts['cell_size']
    model_settings['keep_prob']  = opts['keep_prob']
    model_settings['num_unroll'] = 20 #opts['num_unroll']
    model_settings['learning_rate'] = opts['learning_rate'] #Initial learning rate
    model_settings['l2_regularization'] = opts['l2_regularization']
    model_settings['cell_type'] = opts['cell']
    model_settings['direction'] = opts['direction']
    #print(model_settings)
    
    # TRAIN
    train_settings = dict()
    train_settings['num_step']   = opts['num_step']
    train_settings['batch_size'] = opts['batch_size']
    train_settings['input_dir']  = opts['input_dir']
    train_settings['eval_dir']   = opts['eval_dir']
    train_settings['job_dir']    = opts['job_dir']
    train_settings['job_name']   = opts['job_name']    
    #train_settings['logdir']     = opts['logdir']
    
    if (verbose):
        log_level = logging.INFO
    else:
        log_level = logging.WARNING    
    logging.basicConfig(format='%(levelname)s:%(message)s', level=log_level)
    logging.info('model settings: %s', model_settings)
    logging.info('train settings: %s', train_settings)
    #logging.info('check point dir: %s', ckpoint_dir)
    
    train(model_settings, train_settings)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import argparse
import logging
from .tokenizer import Cutkum

def main():
	p = argparse.ArgumentParser()
	p.add_argument('-v', '--verbose', help='verbose', action='store_true')
	p.add_argument('-m', '--model_file', required=True, help='model file to load')
	g1 = p.add_mutually_exclusive_group(required=True)
	g1.add_argument('-s', '--sentence', help='sentence to parse')
	g1.add_argument('-i', '--input_file', help='input file')
	g1.add_argument('-id', '--input_dir', help='input directory')

	g2 = p.add_mutually_exclusive_group(required=False)
	g2.add_argument('-o', '--output_file', help='output file if --input_file is given')	
	g2.add_argument('-od', '--output_dir', help='output directory if --input_dir is given')

	g3 = p.add_mutually_exclusive_group(required=False)
	g3.add_argument('--max', action='store_const', dest='max_flag', const='max', help='output word boundary using maximum probabilities (default)')
	g3.add_argument('--viterbi', action='store_const', dest='max_flag', const='vit', help='output word boundary using viterbi')
	p.set_defaults(max_flag='max')

	opts = vars(p.parse_known_args()[0])

	verbose = opts['verbose']
	if (verbose):
		log_level = logging.INFO
	else:
		log_level = logging.WARNING    
	logging.basicConfig(format='%(levelname)s:%(message)s', level=log_level)

	# OTHERS ARGS
	model_file		= opts['model_file']
	input_sentence  = opts['sentence']  
	input_file      = opts['input_file']
	output_file     = opts['output_file']
	input_dir       = opts['input_dir']
	output_dir      = opts['output_dir']
	max_flag        = opts['max_flag']

	if (input_file is not None) & (output_file is None):
		p.error('--input_file and --output_file must be given together')
	if (input_dir is not None) & (output_dir is None):
		p.error('--input_dir and --output_dir must be given together')				
	use_viterbi = False
	if (max_flag == 'viterbi'):
		use_viterbi = True

	ck = Cutkum(model_file)
	if input_file is not None:
		ck.tokenize_file(input_file, output_file, use_viterbi)
	elif (input_dir is not None) & (output_dir is not None):
		ck.tokenize_directory(input_dir, output_dir, use_viterbi)
	elif input_sentence is not None:
		words = ck.tokenize(input_sentence, use_viterbi)
		print("|".join(words))

    #graph = load_graph(model_file)

    #with tf.Session(graph=graph) as sess:
    #    char_dict = CharDictionary()
    #    model_settings, model_vars = load_settings(sess)
        
    #    if input_file is not None:
    #        process_input_file(sess, char_dict, model_settings, model_vars, input_file, max_flag)
    #    elif (input_dir is not None) & (output_dir is not None):
    #        process_input_directory(sess, char_dict, model_settings, model_vars, input_dir, output_dir, max_flag)
    #    elif input_sentence is not None:

if __name__ == '__main__':
	main()
    #        process_input_sentence(sess, char_dict, model_settings, model_vars, input_sentence, max_flag)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Pucktada Treeratpituk (https://pucktada.github.io/)
# License: MIT
# 2017-05-21
from __future__ import unicode_literals
from os import listdir
from os.path import isfile, isdir, join
import re
import tempfile
import numpy as np
import tensorflow as tf

from char_dictionary import CharDictionary
from sequence_handler import save_tfrecord, read_and_decode_single_example, get_num_records

# 100 - 1000+
# [71826 43962 14179 12159  3280  1916   950   412   168    67    76]
# 40 - 400+
# [33404 25192 26999 19487 10706  6448  5108  5375  5738  3669  6869]
# total = 132081 sentences
def lines_from_file(filepath):
    '''
        read in a file as a list of words, ex:
        [['กฎหมาย', 'กับ', 'การ', 'เบียดบัง', 'คน', 'จน'], 
         ['จาก', 'ต้นฉบับ', 'เรื่อง', 'คน', 'จน', 'ภาย', 'ใต้', 'กฎหมาย'], 
         ['ไพสิฐ พาณิชย์กุล']]
    '''
    lines = []
    with open(filepath, 'r') as f:
        for s in f: # s is the line string
            if s and (len(s) > 0):
                tokens = re.split('[|]', s.strip())[:-1]
                words = [] # words in a line
                for t in tokens:
                    t = re.sub('<[^>]*>', '', t)
                    words += [t]
                lines += [words]
    return lines

def convert_data_to_tfrecord(dirpaths, outdir):
    '''
       take in the list of directory path, convert them into tf file 
       of sequence of int64 of cids, and labels
    '''
    files = []
    for d in dirpaths:
        files += [(f.replace('.txt', '.tf'), join(d, f)) 
            for f in listdir(d) if isfile(join(d, f))]
    
    char_dict = CharDictionary() # character dictionary 
    
    for fname, fpath in files:
        lines = lines_from_file(fpath)
        
        sequences = []
        label_sequences = []
        for words in lines: # for each line 
            cids, labels = char_dict.words2cids(words)
            sequences += [cids]
            label_sequences += [labels]
        
        tf_filename = join(outdir, fname)
        writer = tf.python_io.TFRecordWriter(tf_filename)
        save_tfrecord(writer, sequences, label_sequences)
        writer.close()

def read_tfrecord_length_distribution(tf_filename):
    num_records = get_num_records(tf_filename)

    buckets = np.zeros(11, dtype=np.int32)
    
    graph = tf.Graph()
    with graph.as_default():
        key, context, features = read_and_decode_single_example([tf_filename], shuffle=True, num_epochs=1)
        cids   = features['source']
        labels = features['target']
    
    with tf.Session(graph=graph) as sess:
        init_op = tf.group(tf.global_variables_initializer(), 
                           tf.local_variables_initializer())
        sess.run(init_op)
    
        coord = tf.train.Coordinator()    
        threads = tf.train.start_queue_runners(coord=coord)

        for i in range(num_records):
            _cids, _labels = sess.run([cids, labels])
            length = len(_cids)
            if length >= 400:
                bid = 10
            else:
                bid = length // 40
            
            buckets[bid] = buckets[bid] + 1
            
        # stop our queue threads and properly close the session
        coord.request_stop()
        coord.join(threads)
    return buckets
        
def read_tfrecord(tf_filename):
    '''
       read in the 'tf_filename' and use CharDictionary to convert
       the sequeneces of cids, labels back to list of words
    '''
    char_dict = CharDictionary()
    
    graph = tf.Graph()
    with graph.as_default():
        key, context, features = read_and_decode_single_example([tf_filename], shuffle=True)
        cids   = features['source']
        labels = features['target']
    
    with tf.Session(graph=graph) as sess:
        init_op = tf.group(tf.global_variables_initializer(), 
                           tf.local_variables_initializer())
        sess.run(init_op)
    
        coord = tf.train.Coordinator()    
        threads = tf.train.start_queue_runners(coord=coord)

        print('\nread back the file...')
        for i in range(100):
            _cids, _labels = sess.run([cids, labels])
            _chars = char_dict.cids2chars(_cids)
            words = char_dict.chars2words(_chars, _labels)
            print(i, words[:10])
            
        # stop our queue threads and properly close the session
        coord.request_stop()
        coord.join(threads)

if __name__ == '__main__':
    dirpaths = ['../data/5M.BEST2010/article', 
                '../data/5M.BEST2010/encyclopedia',
                '../data/5M.BEST2010/news', 
                '../data/5M.BEST2010/novel']
    
    dirname = 'data'
    files = [join(dirname, f) for f in listdir(dirname) if isfile(join(dirname, f))]
            
    buckets = np.zeros(11, dtype=np.int32)
    
    for fname in files:
        print(fname)
        b = read_tfrecord_length_distribution(fname)
        buckets += b
    print(buckets)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import sys
from tensorflow.python.lib.io import file_io
#from .char_dictionary import char_dictionary
from .char_dictionary import CharDictionary
import numpy as np
import re

def fix_space_and_dot(prob_matrix, cids):
    # if prob_matrix is 1D, make it 2D
    if len(np.shape(prob_matrix)) == 1:
        prob_matrix = [prob_matrix]
        
    length = len(prob_matrix)
    assert(length == len(cids))
    for t in range(length):
        # ' '=> 75, '.'=>76
        if (cids[t] == 75): #or (cids[t] == 76):
            for i in range(5):
                prob_matrix[t][i] = 0.0
            prob_matrix[t][4] = 1.0
    return prob_matrix

# shape(prob_matrix) = [#times, #classes]
def max_prob(prob_matrix):
    if len(np.shape(prob_matrix)) == 1:
        prob_matrix = [prob_matrix]
        
    length = len(prob_matrix)

    seq = np.ones(length, 'int32') * 2
    # {B, M, E, S} <=== 1:begin, 2:middle, 3:end, 4:single

    seq[0] = 1
    for t in range(1, length):
        # find where 1, 4 is max
        max_id = np.argmax(prob_matrix[t])
        if (max_id == 1) or (max_id == 4):
            seq[t] = 1

    for t in range(length):
        if seq[t] == 1:
            if (t != length-1) and (seq[t+1] == 1):
                seq[t] = 4
        if seq[t] == 2:
            if (t != length-1) and (seq[t+1] == 1):
                seq[t] = 3
    return seq

# shape(prob_matrix) = [#times, #classes]
def viterbi(prob_matrix):
    """ find the most likely sequence of labels using the viterbi algorithm on prob_matrix """
    TINY          = 1e-6    # to avoid NaNs in logs

    # if prob_matrix is 1D, make it 2D
    if len(np.shape(prob_matrix)) == 1:
        prob_matrix = [prob_matrix]
        
    length = len(prob_matrix)

    probs  = np.zeros_like(prob_matrix)
    backpt = np.ones_like(prob_matrix, dtype=np.int32) * -1

    for i in [0,1,2,3,4]:
        probs[0][i] = np.log(prob_matrix[0][i]+TINY)
        
    # {B, M, E, S} <=== 1:begin, 2:middle, 3:end, 4:single
    for t in range(1, length):
        # E, S -> B | B, M -> M | B, M -> E | E, S -> S
        previous_of = [[0,0], [3,4], [1,2], [1,2], [3,4]]
        for i in range(5):
            prevs = previous_of[i]
            max_id = prevs[np.argmax([probs[t-1][prevs[0]], probs[t-1][prevs[1]]])]
            backpt[t][i] = max_id
            probs[t][i]  = np.log(prob_matrix[t][i]+TINY) + probs[t-1][max_id]

    seq = np.ones(length, 'int32') * -1
    #print(probs[length-1])
    seq[length-1] = np.argmax(probs[length-1])
    #print(seq[length-1])
    max_prob = probs[length-1][seq[length-1]]
    for t in range(1, length):
        seq[length-1-t] = backpt[length-t][seq[length-t]]
    
    return seq

def word_count_from_boundary_array(x):
    n_edges = 0
    for i in x:
        if (i != 0):
            n_edges += 1
    return n_edges-1
    
'''
    words: array of words ... (can have tags and not have tags)
    words = เขา|ร้อง|บท|<POEM>เย็นย่ำ จะค่ำอยู่แล้วลงรอนรอน</POEM>|
    words = เขา|ร้อง|บท|เย็นย่ำ จะค่ำอยู่แล้วลงรอนรอน|
    
    return word boundary array of length+1
    1 is the start of a word
    0 is the middle of a word
    2 is the start of a entity word
'''
def get_boundary_array(words):
    length = 0
    for w in words:
        if w.startswith('<NE>') or w.startswith('<AB>'):
            length += (len(w) - (2*4 + 1))
        elif w.startswith('<NER>'):
            length += (len(w) - (2*5 + 1))
        elif w.startswith('<POEM>'):
            length += (len(w) - (2*6 + 1))
        else:
            length += len(w)
    
    x = np.zeros(length+1, dtype=np.int32)
    x[0] = 1
    
    cur = 0
    for w in words:
        # NE, AB, NER and POEM 
        if w.startswith('<NE>') or w.startswith('<AB>') or w.startswith('<POEM>') or w.startswith('<NER>'):
            x[cur] = 2
            t = re.sub('<[^>]*>', '', w)
            cur += len(t)
            x[cur] = 1
        else:
            cur += len(w)            
            x[cur] = 1

    return x
    
'''
    NE, AB and POEM

                    x= เขา|ร้อง|บท|<POEM>เย็นย่ำ จะค่ำอยู่แล้วลงรอนรอน</POEM>|
    R=4/4, P=4/4    b= เขา|ร้อง|บท|เย็นย่ำ จะค่ำอยู่แล้วลงรอนรอน|
    R=4/4, P=4/4    c= เขา|ร้อง|บท|เย็น|ย่ำ| |จะ|ค่ำ|อยู่แล้ว|ลง|รอนรอน| 11-7=4
    R=2/4, P=2/3    d= เขา|ร้อง|บทเย็น|ย่ำ จะค่ำอยู่แล้วลงรอนรอน|        4-1

       F-Measure = 2 x Precision x Recall / (Precision + Recall)

    Corr       = จำนวนคำที่ผู้เข้าแข่งขันแบ่งได้อย่างถูกต้อง
    RefWord    = จำนวนคำที่แบ่งโดยคณะกรรมการ
    OutputWord = จำนวนคำที่ผู้เข้าแข่งขันแบ่งมาทั้งหมด
    
    x: answer boundary array 
    y: output boundary array
    n_answers: number of words in the answers (before discounting)
    n_outputs: number of words in the outputs (before discounting)
'''
def count_correct_words(answer_boundary, output_boundary, n_answers, n_outputs):
    x = answer_boundary
    y = output_boundary
    
    length = len(x) - 1
    n_outwords = n_outputs
    n_refwords = n_answers
    correct = 0
    
    # counting 'n_outwords'
    #discount = 0
    cur = 0
    while cur <= length:
        if (x[cur] > 1): # start of <NE>, <AB>, <POEM>
            end = cur+1
            while x[end] == 0:
                end += 1
            _c = 0
            # j = [cur,...,end]
            for j in range(cur+1, end+1):
                if y[j] > 0:
                    _c += 1
            n_outwords = n_outwords - (_c - 1)
            if (y[cur] > 1) and (y[end] > 1):
                n_outwords += 1
            #n_outwords 
            cur = end
        else:
            cur += 1
    
    # counting 'correct'
    cur = 0
    while cur <= length:
        # found 
        if (x[cur] > 0) and (y[cur] > 0): # start match
            if x[cur] > 1: # start of <NE>, <AB>, <POEM> & match...
                cur += 1
                # find the end of NE,AB,POEM
                while cur <= length:
                    if x[cur] > 0:
                        break
                    cur += 1
                if y[cur] > 0: # ... also match the ending
                    correct += 1
            else:
                cur += 1            
                while cur <= length:
                    if (x[cur] > 0) and (y[cur] > 0): # end match...
                        correct += 1
                        #print(cur)
                        break
                    elif (x[cur] > 0) or (y[cur] > 0): # not match..
                        break
                    cur += 1
        else:
            cur += 1
    
    return (correct, n_refwords, n_outwords)

'''
    test_files: list of file to load

    return
    
    one_hot_by_t:   [max_time, #examples, #dict]
    seq_lengths:    lengths of each example [#examples], used to feed the model
    chars_mat:      list of list of characters ... can be used for reconstruction of sentences
'''
def load_files_into_matrix(test_files):
    char_dict = CharDictionary()
    input_classes = char_dict.num_char_classes() + 1

    cids_mat  = []
    chars_mat = []
    seq_lengths = []
    for input_file in test_files:
        with file_io.FileIO(input_file, 'r') as f:
            for s in f: # s is the line string
                if (sys.version_info <= (3, 0)):
                    s = s.decode('utf8')
                if s and (len(s) > 0):                    
                    chars = list(s.strip())
                    cids = char_dict.chars2cids(chars)
                    seq_lengths += [len(cids)]
                    cids_mat.append(cids)
                    chars_mat.append(chars)

    n_examples = len(cids_mat)
    
    max_len = max(seq_lengths)
    batch_labels = np.zeros((n_examples, max_len), dtype=np.int32)
    for i in range(n_examples):
        batch_labels[i,0:len(cids_mat[i])] = cids_mat[i]

    in_embedding = np.eye(input_classes) # 2
    one_hot_flat = in_embedding[list(batch_labels.flatten())]
    one_hot      = np.reshape(one_hot_flat, (n_examples, max_len, input_classes))
    
    #(time, batch, in)
    one_hot_by_t = np.transpose(one_hot, (1,0,2))
    return one_hot_by_t, seq_lengths, chars_mat
'''
    test_files: list of file to load

    return
    
    one_hot_by_t:   [max_time, #examples, #dict]
    seq_lengths:    lengths of each example [#examples], used to feed the model
    chars_mat:      list of list of characters ... can be used for reconstruction of sentences
    boundary_mat:   list of boundary array (each is of length #chars+1)
'''
def load_validation_set(test_files):
    char_dict = CharDictionary()
    input_classes = char_dict.num_char_classes() + 1

    cids_mat  = []
    chars_mat = []
    boundary_mat = []
    seq_lengths = []
    for input_file in test_files:
        with file_io.FileIO(input_file, 'r') as f:
            for s in f: # s is the line string
                if (sys.version_info <= (3, 0)):
                    s = s.decode('utf8')
                if s and (len(s) > 0):
                    t = re.sub("<[^>]*>", "", s.strip())
                    t = re.sub("[|]", "", t)

                    # [:-1] remove the ending empty word (resulting from "|" at the end of line in BEST2010)
                    answers = re.split('[|]', s.strip())[:-1]
                    boundary = get_boundary_array(answers)
                    boundary_mat.append(boundary)
                    
                    chars = list(t)
                    cids = char_dict.chars2cids(chars)
                    seq_lengths += [len(cids)]
                    cids_mat.append(cids)
                    chars_mat.append(chars)

    n_examples = len(cids_mat)
    
    max_len = max(seq_lengths)
    batch_labels = np.zeros((n_examples, max_len), dtype=np.int32)
    for i in range(n_examples):
        batch_labels[i,0:len(cids_mat[i])] = cids_mat[i]

    in_embedding = np.eye(input_classes) # 2
    one_hot_flat = in_embedding[list(batch_labels.flatten())]
    one_hot      = np.reshape(one_hot_flat, (n_examples, max_len, input_classes))
    
    #(time, batch, in)
    one_hot_by_t = np.transpose(one_hot, (1,0,2))
    
    return one_hot_by_t, seq_lengths, chars_mat, boundary_mat

if __name__ == '__main__':
    

    #test_files = ['data/test_txt/article_00179.txt', 
    #    'data/test_txt/encyclopedia_00099.txt', 
    #    'data/test_txt/news_00088.txt', 
    #    'data/test_txt/novel_00098.txt']
    test_files = ['data/test_txt/article_00179.txt']
        
    one_hot_by_t, seq_lengths, chars_mat, boundary_mat = load_validation_set(test_files)
    print(one_hot_by_t.shape)
    #print(len(seq_lengths))
    #print(len(chars_mat))
    #print(len(boundary_mat))
    #labels = viterbi(probs)
    #words  = char_dict.chars2words(chars, labels)
    #print('|'.join(words))
    
    answers = u"เขา|ร้อง|บท|<POEM>เย็นย่ำ จะค่ำอยู่แล้วลงรอนรอน</POEM>|".split('|')[:-1]
    print('|'.join(answers))

    words = u"เขา|ร้อง|บท|เย็น|ย่ำ| |จะ|ค่ำ|อยู่แล้ว|ลง|รอนรอน".split('|')
    print('|'.join(words))

    x = get_boundary_array(answers)
    y = get_boundary_array(words)
    print(x)
    print(y)
    print(len(answers))
    print(len(words))
    print('len x:', word_count_from_boundary_array(x))
    print('len y:', word_count_from_boundary_array(y))    
    correct, n_refwords, n_outwords = count_correct_words(x, y, len(answers), len(words))
    
    r = float(correct) / float(n_refwords)
    p = float(correct) / float(n_outwords)
    f = 2 * p * r / (p + r)
    print("recall = %.3f, precision = %.3f, F = %.3f" % (r, p, f))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Pucktada Treeratpituk (https://pucktada.github.io/)
# License: MIT
# 2017-05-21
from __future__ import unicode_literals
from __future__ import print_function
import numpy as np
import sys

class CharDictionary:
    """ a utility class for converting characters into character ids. 
        the id 0 is reserved for padding (both for character ids and label ids)
    """
    
    def __init__(self):
        thai_chars = 'กขฃคฅฆงจฉชซฌญฐฎฏฐฑฒณดตถทธนบปผฝพฟภมยรฤลฦวศษสหฬอฮฯะัาำิีึืุูเแโใไๅๆ็่้๊๋์ํ' # 1..73
        numbers = '๐๑๒๓๔๕๖๗๘๙0123456789' # => 74 ' '=> 75, '.'=>76
        symbols = '()[]<>{}:;^+-*/_=#!~%\\/\'"`?' # 77
        eng_chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ' # 78

        self._class_for = {'B':1, 'M':2, 'E':3, 'S':4}

        self._cid2char = ['', '_'] # ''=padding, _=_UNK
        self._char2cid = dict()
        
        next_idx = 2
        # 1 ... len(thai_chars)
        for c in thai_chars:
            self._char2cid[c] = next_idx
            self._cid2char.append(c)
            next_idx += 1
                    
        last_char_id = len(thai_chars) + 1

        for d in numbers:
            self._char2cid[d] = last_char_id + 1
        self._cid2char.append('$')
        
        self._char2cid[' '] = last_char_id + 2
        self._cid2char.append(' ')
        
        self._char2cid['.'] = last_char_id + 3
        self._cid2char.append('.')
        
        for s in symbols:
            self._char2cid[s] = last_char_id + 4
        self._cid2char.append('#')  

        for s in eng_chars:
            self._char2cid[s] = last_char_id + 5
        self._cid2char.append('A')

    def num_char_classes(self):
        """ return the number of character classes """
        # minus 1... cause cid 0 is not used... reserved for padding
        return len(self._cid2char) - 1 # 
    
    def num_label_classes(self):
        """ return the number of label classes """        
        return len(self._class_for) # B, M, E, S

    def words2cids(self, words):
        """ return the arrays of character ids and labels for a given array of words        
        """
        chars, labels = self.words2chars(words)
        cids = self.chars2cids(chars)
        return cids, labels
        
    # labels = {B, M, E, S} <=== 0:begin, 1:middle, 2:end, 3:single
    # convert words array to character array, and create a labels
    def words2chars(self, words):
        """ return the arrays of characters and labels for a given array of words        
        """
        chars, labels = [], []

        for w in words:
            if len(w) == 1: # single char
                chars.append(w[0])
                labels.append(self._class_for['S']) # 'S'
            elif len(w) > 0:
                chars.append(w[0])
                labels.append(self._class_for['B']) # 'B'
                for i in range(len(w)-2):
                    chars.append(w[i+1])
                    labels.append(self._class_for['M']) # 'M'
                chars.append(w[-1])
                labels.append(self._class_for['E']) # 'E'
        return chars, labels

    def chars2words(self, chars, labels):
        """ return array of words for the given arrays of characters and labels
            (the reverse of 'words2chars')
        """

        words = []
        word = ''
        class_for = {'B':1, 'M':2, 'E':3, 'S':4}
        for i in range(len(chars)):
            c = chars[i] # reverse_dict[data[i]]
            l = labels[i] # 0:begin, 1:middle, 2:end, 3:single
            if l == self._class_for['B']: # begin
                if len(word) != 0: # somehow, still have left-over word
                    words += [word]
                word = c
            elif l == self._class_for['M']: # middle
                word += c
            elif l == self._class_for['E']: # end
                word += c
                words += [word]
                word = ''
            elif l == self._class_for['S']: # single
                if (len(word) != 0):
                    words += [word]
                words += [c]
                word = ''
        if (len(word) != 0):
            words += [word]
        return words
    
    def cidOf(self, c):
        """ return the character id of the given character 'c' """
        if c in self._char2cid:
            return self._char2cid[c]
        else:
            return 1 # UNK_ID
    
    def chars2cids(self, chars):
        """ return the array of character ids for the given arrays of characters """        
        return [self.cidOf(c) for c in chars]
        
    def cids2chars(self, cids):
        """ return the array of characters for the given arrays of character ids 
            (the reverse of chars2cids)
        """
        return [self._cid2char[i] for i in cids]

def test_char_dict():
    print('testing character dictionary')
    words = ['กฎหมาย', 'กับ', '1', 'การ', 'เบียดบัง', 'คน']
    
    dic = CharDictionary()
    
    sen = list('สารานุกรมไทยสำหรับเยาวชนฯ')
    chars = dic.words2chars(sen)
    print(chars)
    
    chars, labels = dic.words2chars(words)
    cids = dic.chars2cids(chars)
    
    print(' '.join(chars))
    print(dic.chars2cids(chars))
    print(' '.join(dic.cids2chars(cids)))

if __name__ == '__main__':
    test_char_dict()


from setuptools import find_packages
from setuptools import setup

REQUIRED_PACKAGES = ['numpy==1.13', 'ConfigArgParse>=0.12', 'tensorflow']

def readme():
	with open('README.md') as f:
		return f.read()

setup(name='cutkum',
	version='1.4',
	description='Thai Word-Segmentation with LSTM in Tensorflow',
	long_description=readme(),
	keywords='funniest joke comedy flying circus',
	url='https://github.com/pucktada/cutkum',
	author='Puck Treeratpituk',
	author_email='pucktada@gmail.com',
	license='MIT',
    packages=find_packages(),
    install_requires=REQUIRED_PACKAGES,
	entry_points={
		'console_scripts': ['cutkum=cutkum.command_line:main'],
	},
	include_package_data=True,
	zip_safe=True)



#!/usr/bin/env python
#
# Pucktada Treeratpituk (https://pucktada.github.io/)
# License: MIT
# 2017-05-01
#
# A recurrent neural network model (LSTM) for thai word segmentation
import logging
import re
import numpy as np
import tensorflow as tf
import tensorflow.contrib.layers as layers
import tensorflow.contrib.rnn as rnn

from . import char_dictionary

def load_settings(sess):

    model_settings = dict()        
    model_vars = dict()

    graph = tf.get_default_graph()

    configs = ['cell_sizes', 'num_layers', 'input_classes', 'label_classes', 'learning_rate', 'l2_regularization', 'cell_type', 'direction']
    for c in configs:
        name = 'prefix/%s:0' % c
        model_settings[c] = sess.run(graph.get_tensor_by_name(name))

    model_vars['inputs']  = graph.get_tensor_by_name('prefix/placeholder/inputs:0')
    model_vars['seq_lengths'] = graph.get_tensor_by_name('prefix/placeholder/seq_lengths:0')
    model_vars['keep_prob'] = graph.get_tensor_by_name('prefix/placeholder/keep_prob:0')

    model_vars['probs'] = graph.get_tensor_by_name('prefix/batch_probs:0')

    return model_settings, model_vars

def load_graph(model_file):
    """ loading necessary configuration of the network from the meta file & 
        the checkpoint file together with variables that are needed for the inferences
    """

    # We load the protobuf file from the disk and parse it to retrieve the 
    # unserialized graph_def
    with tf.gfile.GFile(model_file, "rb") as f:
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(f.read())

    # Then, we can use again a convenient built-in function to import a graph_def into the 
    # current default Graph
    with tf.Graph().as_default() as graph:
        tf.import_graph_def(
            graph_def,
            input_map=None, 
            return_elements=None, 
            name='prefix',
            op_dict=None,
            producer_op_list=None)

    return graph

def load_model2(sess, meta_file, checkpoint_file):
    """ loading necessary configuration of the network from the meta file & 
        the checkpoint file together with variables that are needed for the inferences
    """
    saver = tf.train.import_meta_graph(meta_file, clear_devices=True)
    saver.restore(sess, checkpoint_file)    

    configs = tf.get_collection('configs')
    pvars   = tf.get_collection('placeholders')

    model_settings = dict()
    for c in configs:
        name = c.name.split(':')[0]
        model_settings[name] = sess.run(c)
        
    model_vars = dict()
    for p in pvars:
        scope, name, _ = re.split('[:/]', p.name)
        model_vars[name] = p
    model_vars['probs'] = tf.get_collection('probs')[0]

    return model_settings, model_vars

class CkModel:
    """ cutkum model: LSTM recurrent neural network model """
    
    def __init__(self, model_settings):
        logging.info('...init WordSegmentor')
                
        self.cell_sizes = model_settings["cell_size"] # list of cell_size, same length as num_layers
        #self.num_unroll = model_settings["num_unroll"]
        self.num_layers = model_settings["num_layers"]        
        self.input_classes = model_settings['input_classes']
        self.label_classes = model_settings['label_classes']
        self.learning_rate = model_settings['learning_rate']
        self.l2_regularization = model_settings['l2_regularization'] # 0.1
        self.cell_type = model_settings['cell_type']
        self.direction = model_settings['direction']
        
        tf.add_to_collection('configs', tf.constant(self.cell_sizes, name="cell_sizes"))
        #tf.add_to_collection('configs', tf.constant(self.num_unroll, name="num_unroll"))
        tf.add_to_collection('configs', tf.constant(self.num_layers, name="num_layers"))
        tf.add_to_collection('configs', tf.constant(self.input_classes, name="input_classes"))
        tf.add_to_collection('configs', tf.constant(self.label_classes, name="label_classes"))
        tf.add_to_collection('configs', tf.constant(self.learning_rate, name="learning_rate"))           
        tf.add_to_collection('configs', tf.constant(self.l2_regularization, name="l2_regularization"))
        tf.add_to_collection('configs', tf.constant(self.cell_type, name="cell_type"))
        tf.add_to_collection('configs', tf.constant(self.direction, name="direction"))
        
        self.global_step = tf.Variable(0, dtype=tf.int32, trainable=False, name='global_step')
        self.increment_global_step_op = tf.assign(self.global_step, self.global_step+1)
        
    def _create_placeholders(self):
        logging.info('...create placeholder')
        with tf.name_scope("placeholder"):       
            # (time, batch, in)        
            self.inputs  = tf.placeholder(tf.float32, (None, None, self.input_classes), name="inputs")
            # (time, batch, out)
            self.outputs = tf.placeholder(tf.float32, (None, None, self.label_classes), name="outputs")
            # [batch]
            self.seq_lengths = tf.placeholder(tf.int32, [None], name="seq_lengths")
        
            #self.fw_state = tf.placeholder(tf.float32, [self.num_layers, 2, None, self.cell_sizes], name="fw_state")
            #self.bw_state = tf.placeholder(tf.float32, [self.num_layers, 2, None, self.cell_sizes], name="bw_state")
            self.keep_prob = tf.placeholder(tf.float32, name="keep_prob")
            
            tf.add_to_collection('placeholders', self.inputs)
            tf.add_to_collection('placeholders', self.outputs)
            tf.add_to_collection('placeholders', self.seq_lengths)        
            #tf.add_to_collection('placeholders', self.fw_state)
            #tf.add_to_collection('placeholders', self.bw_state)
            tf.add_to_collection('placeholders', self.keep_prob)         
    
    def _inference(self):
        logging.info('...create inference')
        
        #fw_state_list = tf.unstack(self.fw_state, axis=0)
        #fw_state_tuple = tuple(
        #    [tf.contrib.rnn.LSTMStateTuple(fw_state_list[idx][0], fw_state_list[idx][1])
        #     for idx in range(self.num_layers)])

        #bw_state_list = tf.unstack(self.bw_state, axis=0)
        #bw_state_tuple = tuple(
        #    [tf.contrib.rnn.LSTMStateTuple(bw_state_list[idx][0], bw_state_list[idx][1])
        #     for idx in range(self.num_layers)])
             
        fw_cells = list()
        for i in range(0, self.num_layers):
            if (self.cell_type == 'lstm'):
                cell = rnn.LSTMCell(num_units=self.cell_sizes[i], state_is_tuple=True)                
            elif (self.cell_type == 'gru'):
                # change to GRU
                cell = rnn.LSTMCell(num_units=self.cell_sizes[i], state_is_tuple=True)
            else:
                cell = rnn.BasicRNNCell(num_units=self.cell_sizes[i])
            cell = rnn.DropoutWrapper(cell, output_keep_prob=self.keep_prob)
            fw_cells.append(cell)        
        self.fw_cells = rnn.MultiRNNCell(fw_cells, state_is_tuple=True)
        
        if (self.direction == 2): # bidirectional
            print('bidirectional')
            bw_cells = list()
            for i in range(0, self.num_layers):
                if (self.cell_type == 'lstm'):
                    cell = rnn.LSTMCell(num_units=self.cell_sizes[i], state_is_tuple=True)
                elif (self.cell_type == 'gru'):
                    # change to GRU
                    cell = rnn.LSTMCell(num_units=self.cell_sizes[i], state_is_tuple=True)
                else:
                    cell = rnn.BasicRNNCell(num_units=self.cell_sizes[i])
                cell = rnn.DropoutWrapper(cell, output_keep_prob=self.keep_prob)
                bw_cells.append(cell)        
            self.bw_cells = rnn.MultiRNNCell(bw_cells, state_is_tuple=True)
        
        if (self.direction == 1):
            rnn_outputs, states = tf.nn.dynamic_rnn(
                self.fw_cells, 
                self.inputs, 
                #initial_state=fw_state_tuple,
                sequence_length=self.seq_lengths,
                dtype=tf.float32, time_major=True)
        else: # self.direction = 2        
            # bidirectional rnn
            outputs, states  = tf.nn.bidirectional_dynamic_rnn(
                cell_fw=self.fw_cells,
                cell_bw=self.bw_cells,
                #initial_state_fw=fw_state_tuple,
                #initial_state_bw=bw_state_tuple,
                dtype=tf.float32,
                sequence_length=self.seq_lengths, 
                inputs=self.inputs, time_major=True)
            rnn_outputs = tf.concat(outputs, axis=2)

        # project output from rnn output size to OUTPUT_SIZE. Sometimes it is worth adding
        # an extra layer here.
        self.projection = lambda x: layers.linear(x, 
            num_outputs=self.label_classes, activation_fn=tf.nn.sigmoid)

        self.logits = tf.map_fn(self.projection, rnn_outputs, name="logits")
        self.probs  = tf.nn.softmax(self.logits, name="probs")
        self.states = states

        tf.add_to_collection('probs',  self.probs)
    
    def _create_loss(self):
        logging.info('...create loss')
        with tf.name_scope("loss"):
            # shape=[Time * Batch, label_classes]   
            outputs_flat = tf.reshape(self.outputs, [-1, self.label_classes])
            logits_flat  = tf.reshape(self.logits,  [-1, self.label_classes])

            # calculate the losses shape=[Time * Batch]
            losses = tf.nn.softmax_cross_entropy_with_logits(
                labels=outputs_flat, logits=logits_flat)
            
             # create mask [Time * Batch] where 0: padded, 1: not-padded
            mask = outputs_flat[:,0]
            mask = tf.abs(tf.subtract(mask, tf.ones_like(mask)))
            # mask the losses
            masked_losses = mask * losses

            l2_reg = self.l2_regularization
            l2 = l2_reg * sum(tf.nn.l2_loss(tf_var) for tf_var in tf.trainable_variables() 
                if not ("noreg" in tf_var.name or "Bias" in tf_var.name))
            self.losses = masked_losses + l2

            self.num_entries = tf.reduce_sum(mask)
            self.mean_loss = tf.reduce_sum(masked_losses) / self.num_entries
            
            # accuracy
            correct_pred = tf.cast(tf.equal(tf.argmax(outputs_flat, 1), tf.argmax(logits_flat, 1)), tf.float32)
            mask_correct_pred = mask * correct_pred
            self.accuracy = tf.reduce_sum(mask_correct_pred) / self.num_entries
        
    def _create_optimizer(self):
        logging.info('...create optimizer')
        with tf.name_scope("train"):        
            #self.optimizer = tf.train.AdamOptimizer(self.learning_rate).minimize(self.mean_loss, global_step=self.global_step)
            max_gradient_norm = 1.0
            params    = tf.trainable_variables()
            gradients = tf.gradients(self.mean_loss, params)
            clipped_gradients, norm = tf.clip_by_global_norm(gradients, max_gradient_norm)
            #self.train_op = tf.train.AdamOptimizer(learning_rate=self.learning_rate)\
            #    .apply_gradients(zip(clipped_gradients, params), global_step=self.global_step)
            self.train_op = tf.train.AdamOptimizer(learning_rate=self.learning_rate)\
                .apply_gradients(zip(clipped_gradients, params))
    
    def _create_summary(self):
        logging.info('...create summary')
        tf.summary.scalar("mean_loss", self.mean_loss)
        tf.summary.scalar("accuracy",  self.accuracy)
        self.summary_op = tf.summary.merge_all()
    
    def build_graph(self):
        self._create_placeholders()
        self._inference()
        self._create_loss()
        self._create_optimizer()
        self._create_summary()

if __name__ == '__main__':
    print('create word segmentor model')
    char_dict = CharDictionary()
    
    # MODEL
    model_settings = dict()
    #model_settings["l2_regularisation"] = 0.0 # not usring right now
    model_settings['num_unroll'] = 12
    model_settings['num_layers'] = 3
    model_settings['cell_size'] = 64
    model_settings['input_classes'] = char_dict.num_char_classes() + 1
    model_settings['label_classes'] = char_dict.num_label_classes() + 1
    model_settings['learning_rate'] = 0.001 # Initial learning rate
    
    model = CkModel(model_settings)
    model.build_graph()



# -*- coding: utf-8 -*-
#
# Pucktada Treeratpituk (https://pucktada.github.io/)
# License: MIT
# 2017-05-01
#
# A command-line function to load a trained model and compute the word segmentation
from __future__ import unicode_literals
from os import listdir
from os.path import isfile, isdir, join
import logging
import math
import numpy as np
import re
import sys
import tensorflow as tf
from .char_dictionary import CharDictionary
from .ck_model import CkModel, load_graph, load_settings
from .util import max_prob, viterbi, load_files_into_matrix, load_validation_set

class Cutkum:
	""" cutkum model: LSTM recurrent neural network model """

	def __init__(self, model_file):
		logging.info('...init Cutkum')

		self.char_dict = CharDictionary()
		#self.model_file = 'model/lstm.l6.d2.pb'
		self.graph = load_graph(model_file)

		with tf.Session(graph=self.graph) as sess:
			model_settings, model_vars = load_settings(sess)
			self.model_settings = model_settings
			self.model_vars = model_vars

		logging.info('...done init')

	def _process_sentence(self, one_hot_by_t):
		""" run the inference to segment the given 'sentence' represented by the one_hot vector by time
		"""
		probs = None
		with tf.Session(graph=self.graph) as sess:
			feed = { 
				self.model_vars['inputs']:  one_hot_by_t, 
				self.model_vars['seq_lengths']: [one_hot_by_t.shape[0]], 
				self.model_vars['keep_prob']: 1.0
			}
			probs = sess.run([self.model_vars['probs']], feed_dict=feed)

		return np.squeeze(probs)

	def _process_batch(self, one_hot_by_t, seq_lengths):

		prob_list = []
		with tf.Session(graph=self.graph) as sess:
			for i in range(len(seq_lengths)): # for each line
				if (i % 10) == 0:
					logging.info('sen: %d', i)
				if (seq_lengths[i] != 0):
					line_one_hot_by_t = one_hot_by_t[0:seq_lengths[i],i:(i+1),:]
					feed = { self.model_vars['inputs']: line_one_hot_by_t, 
						self.model_vars['seq_lengths']: [seq_lengths[i]], 
						self.model_vars['keep_prob']: 1.0
					}
					probs = sess.run(self.model_vars['probs'], feed_dict=feed)
					prob_list.append(probs)
				else:
					prob_list.append([])
		return prob_list

	def tokenize(self, sentence, use_viterbi=False):
		""" run the inference to segment the given 'sentence' into words seperated by '|' 
		"""

		if (sys.version_info > (3, 0)):
			uni_s = u"%s" % sentence # only for 2.7 (for Python 3, no need to decode)
		else:
			uni_s = u"%s" % sentence.decode('utf-8') # only for 2.7 (for Python 3, no need to decode)

		chars = list(uni_s.strip())
		cids  = self.char_dict.chars2cids(chars)

		in_embedding = np.eye(self.model_settings['input_classes'])
		one_hot      = in_embedding[cids]
		one_hot_by_t = np.expand_dims(one_hot, 1)

		p  = self._process_sentence(one_hot_by_t)
		if (use_viterbi): # viterbi
			labels = viterbi(p)
		else:
			labels = max_prob(p)

		words  = self.char_dict.chars2words(chars, labels)
		return words

	def tokenize_file(self, input_file, output_file, use_viterbi=False):
		""" run the inference to segment the given file into words seperated by '|' 
		"""

		# load file into matrix
		one_hot_by_t, seq_lengths, chars_mat = load_files_into_matrix([input_file])
		probs = self._process_batch(one_hot_by_t, seq_lengths)

		with open(output_file, 'w') as outfile:

			n_examples= len(probs)
			for i in range(n_examples):
				if (seq_lengths[i] != 0):
					p = probs[i][0:seq_lengths[i], 0, :]

					if (use_viterbi): # viterbi
						labels = viterbi(p)
					else:
						labels = max_prob(p)

					words  = self.char_dict.chars2words(chars_mat[i], labels)
					line_out = '|'.join(words)
					if (sys.version_info <= (3, 0)):
						line_out = line_out.encode('utf8')
					outfile.write(line_out)
					outfile.write('\n')
				else:
					outfile.write('\n')

	def tokenize_directory(self, input_dir, output_dir, use_viterbi=False):
		""" run the inference to segment each file in the given directory into words seperated by '|' 
		"""
		input_files = listdir(input_dir)

		for f in input_files:
			infile_path  = join(input_dir, f)
			outfile_path = join(output_dir, f)

			if isfile(outfile_path):
				logging.info('%s ... SKIP (already exists in the output directory)', f)
			else:
				outfile = open(outfile_path, 'w')
				one_hot_by_t, seq_lengths, chars_mat = load_files_into_matrix([infile_path])
				probs = self._process_batch(one_hot_by_t, seq_lengths)

				n_examples= len(probs)
				for i in range(n_examples):
					if (seq_lengths[i] != 0):
						p = probs[i][0:seq_lengths[i], 0, :]

						if (use_viterbi): # viterbi
							labels = viterbi(p)
						else:
							labels = max_prob(p)
						words  = self.char_dict.chars2words(chars_mat[i], labels)

						line_out = '|'.join(words)
						if (sys.version_info <= (3, 0)):
							line_out = line_out.encode('utf8')
						outfile.write(line_out)
						outfile.write('\n')
					else:
						outfile.write('\n')
				outfile.close()
				logging.info('%s ... DONE', f)


